@@ -7,11 +7,11 @@
  *		A big hack of the regexp.c code!! Contributed by
  *		Keith Parks <emkxp01@mtcc.demon.co.uk> (7/95).
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	$Header: /home/projects/pgsql/cvsroot/pgsql/src/backend/utils/adt/like.c,v 1.45 2001/03/22 03:59:51 momjian Exp $
+ *	$Header: /cvsroot/pgsql/src/backend/utils/adt/like.c,v 1.53 2002/09/03 21:45:42 petere Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -19,9 +19,7 @@
 
 #include <ctype.h>
 
-#ifdef MULTIBYTE
 #include "mb/pg_wchar.h"
-#endif
 #include "utils/builtins.h"
 
 
@@ -34,9 +32,16 @@ static int MatchText(unsigned char *t, int tlen,
 		  unsigned char *p, int plen);
 static int MatchTextIC(unsigned char *t, int tlen,
 			unsigned char *p, int plen);
+static int MatchBytea(unsigned char *t, int tlen,
+		   unsigned char *p, int plen);
+static text *do_like_escape(text *, text *);
 
+static int MBMatchText(unsigned char *t, int tlen,
+			unsigned char *p, int plen);
+static int MBMatchTextIC(unsigned char *t, int tlen,
+			  unsigned char *p, int plen);
+static text *MB_do_like_escape(text *, text *);
 
-#ifdef MULTIBYTE
 /*--------------------
  * Support routine for MatchText. Compares given multibyte streams
  * as wide characters. If they match, returns 1 otherwise returns 0.
@@ -88,7 +93,7 @@ iwchareq(unsigned char *p1, unsigned char *p2)
 		return (0);
 
 	/*
-	 * ok, p1 and p2 are both > CHARMAX, then they must be multi-byte
+	 * ok, p1 and p2 are both > CHARMAX, then they must be multibyte
 	 * characters
 	 */
 	l = pg_mblen(p1);
@@ -100,9 +105,6 @@ iwchareq(unsigned char *p1, unsigned char *p2)
 	return (c1[0] == c2[0]);
 }
 
-#endif
-
-#ifdef MULTIBYTE
 #define CHAREQ(p1, p2) wchareq(p1, p2)
 #define ICHAREQ(p1, p2) iwchareq(p1, p2)
 #define NextChar(p, plen) \
@@ -113,13 +115,28 @@ iwchareq(unsigned char *p1, unsigned char *p2)
 		 while (__l-- > 0) \
 			 *(dst)++ = *(src)++; \
 	   } while (0)
-#else
+
+#define MatchText	MBMatchText
+#define MatchTextIC MBMatchTextIC
+#define do_like_escape	MB_do_like_escape
+#include "like_match.c"
+#undef CHAREQ
+#undef ICHAREQ
+#undef NextChar
+#undef CopyAdvChar
+#undef MatchText
+#undef MatchTextIC
+#undef do_like_escape
+
 #define CHAREQ(p1, p2) (*(p1) == *(p2))
 #define ICHAREQ(p1, p2) (tolower(*(p1)) == tolower(*(p2)))
 #define NextChar(p, plen) ((p)++, (plen)--)
 #define CopyAdvChar(dst, src, srclen) (*(dst)++ = *(src)++, (srclen)--)
-#endif
 
+#define BYTEA_CHAREQ(p1, p2) (*(p1) == *(p2))
+#define BYTEA_NextChar(p, plen) ((p)++, (plen)--)
+#define BYTEA_CopyAdvChar(dst, src, srclen) (*(dst)++ = *(src)++, (srclen)--)
+#include "like_match.c"
 
 /*
  *	interface routines called by the function manager
@@ -141,7 +158,10 @@ namelike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchText(s, slen, p, plen) == LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchText(s, slen, p, plen) == LIKE_TRUE);
+	else
+		result = (MBMatchText(s, slen, p, plen) == LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -162,7 +182,10 @@ namenlike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchText(s, slen, p, plen) != LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchText(s, slen, p, plen) != LIKE_TRUE);
+	else
+		result = (MBMatchText(s, slen, p, plen) != LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -183,7 +206,10 @@ textlike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchText(s, slen, p, plen) == LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchText(s, slen, p, plen) == LIKE_TRUE);
+	else
+		result = (MBMatchText(s, slen, p, plen) == LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -204,7 +230,52 @@ textnlike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchText(s, slen, p, plen) != LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchText(s, slen, p, plen) != LIKE_TRUE);
+	else
+		result = (MBMatchText(s, slen, p, plen) != LIKE_TRUE);
+
+	PG_RETURN_BOOL(result);
+}
+
+Datum
+bytealike(PG_FUNCTION_ARGS)
+{
+	bytea	   *str = PG_GETARG_BYTEA_P(0);
+	bytea	   *pat = PG_GETARG_BYTEA_P(1);
+	bool		result;
+	unsigned char *s,
+			   *p;
+	int			slen,
+				plen;
+
+	s = VARDATA(str);
+	slen = (VARSIZE(str) - VARHDRSZ);
+	p = VARDATA(pat);
+	plen = (VARSIZE(pat) - VARHDRSZ);
+
+	result = (MatchBytea(s, slen, p, plen) == LIKE_TRUE);
+
+	PG_RETURN_BOOL(result);
+}
+
+Datum
+byteanlike(PG_FUNCTION_ARGS)
+{
+	bytea	   *str = PG_GETARG_BYTEA_P(0);
+	bytea	   *pat = PG_GETARG_BYTEA_P(1);
+	bool		result;
+	unsigned char *s,
+			   *p;
+	int			slen,
+				plen;
+
+	s = VARDATA(str);
+	slen = (VARSIZE(str) - VARHDRSZ);
+	p = VARDATA(pat);
+	plen = (VARSIZE(pat) - VARHDRSZ);
+
+	result = (MatchBytea(s, slen, p, plen) != LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -229,7 +300,10 @@ nameiclike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchTextIC(s, slen, p, plen) == LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchTextIC(s, slen, p, plen) == LIKE_TRUE);
+	else
+		result = (MBMatchTextIC(s, slen, p, plen) == LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -250,7 +324,10 @@ nameicnlike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchTextIC(s, slen, p, plen) != LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchTextIC(s, slen, p, plen) != LIKE_TRUE);
+	else
+		result = (MBMatchTextIC(s, slen, p, plen) != LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -271,7 +348,10 @@ texticlike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchTextIC(s, slen, p, plen) == LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchTextIC(s, slen, p, plen) == LIKE_TRUE);
+	else
+		result = (MBMatchTextIC(s, slen, p, plen) == LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -292,7 +372,10 @@ texticnlike(PG_FUNCTION_ARGS)
 	p = VARDATA(pat);
 	plen = (VARSIZE(pat) - VARHDRSZ);
 
-	result = (MatchTextIC(s, slen, p, plen) != LIKE_TRUE);
+	if (pg_database_encoding_max_length() == 1)
+		result = (MatchTextIC(s, slen, p, plen) != LIKE_TRUE);
+	else
+		result = (MBMatchTextIC(s, slen, p, plen) != LIKE_TRUE);
 
 	PG_RETURN_BOOL(result);
 }
@@ -307,6 +390,25 @@ like_escape(PG_FUNCTION_ARGS)
 	text	   *pat = PG_GETARG_TEXT_P(0);
 	text	   *esc = PG_GETARG_TEXT_P(1);
 	text	   *result;
+
+	if (pg_database_encoding_max_length() == 1)
+		result = do_like_escape(pat, esc);
+	else
+		result = MB_do_like_escape(pat, esc);
+
+	PG_RETURN_TEXT_P(result);
+}
+
+/*
+ * like_escape_bytea() --- given a pattern and an ESCAPE string,
+ * convert the pattern to use Postgres' standard backslash escape convention.
+ */
+Datum
+like_escape_bytea(PG_FUNCTION_ARGS)
+{
+	bytea	   *pat = PG_GETARG_BYTEA_P(0);
+	bytea	   *esc = PG_GETARG_BYTEA_P(1);
+	bytea	   *result;
 	unsigned char *p,
 			   *e,
 			   *r;
@@ -328,7 +430,6 @@ like_escape(PG_FUNCTION_ARGS)
 
 	if (elen == 0)
 	{
-
 		/*
 		 * No escape character is wanted.  Double any backslashes in the
 		 * pattern to make them act like ordinary characters.
@@ -337,16 +438,15 @@ like_escape(PG_FUNCTION_ARGS)
 		{
 			if (*p == '\\')
 				*r++ = '\\';
-			CopyAdvChar(r, p, plen);
+			BYTEA_CopyAdvChar(r, p, plen);
 		}
 	}
 	else
 	{
-
 		/*
 		 * The specified escape must be only a single character.
 		 */
-		NextChar(e, elen);
+		BYTEA_NextChar(e, elen);
 		if (elen != 0)
 			elog(ERROR, "ESCAPE string must be empty or one character");
 		e = VARDATA(esc);
@@ -357,7 +457,7 @@ like_escape(PG_FUNCTION_ARGS)
 		if (*e == '\\')
 		{
 			memcpy(result, pat, VARSIZE(pat));
-			PG_RETURN_TEXT_P(result);
+			PG_RETURN_BYTEA_P(result);
 		}
 
 		/*
@@ -368,10 +468,10 @@ like_escape(PG_FUNCTION_ARGS)
 		afterescape = false;
 		while (plen > 0)
 		{
-			if (CHAREQ(p, e) && !afterescape)
+			if (BYTEA_CHAREQ(p, e) && !afterescape)
 			{
 				*r++ = '\\';
-				NextChar(p, plen);
+				BYTEA_NextChar(p, plen);
 				afterescape = true;
 			}
 			else if (*p == '\\')
@@ -379,12 +479,12 @@ like_escape(PG_FUNCTION_ARGS)
 				*r++ = '\\';
 				if (!afterescape)
 					*r++ = '\\';
-				NextChar(p, plen);
+				BYTEA_NextChar(p, plen);
 				afterescape = false;
 			}
 			else
 			{
-				CopyAdvChar(r, p, plen);
+				BYTEA_CopyAdvChar(r, p, plen);
 				afterescape = false;
 			}
 		}
@@ -392,58 +492,14 @@ like_escape(PG_FUNCTION_ARGS)
 
 	VARATT_SIZEP(result) = r - ((unsigned char *) result);
 
-	PG_RETURN_TEXT_P(result);
+	PG_RETURN_BYTEA_P(result);
 }
 
-
 /*
-**	Originally written by Rich $alz, mirror!rs, Wed Nov 26 19:03:17 EST 1986.
-**	Rich $alz is now <rsalz@bbn.com>.
-**	Special thanks to Lars Mathiesen <thorinn@diku.dk> for the LABORT code.
-**
-**	This code was shamelessly stolen from the "pql" code by myself and
-**	slightly modified :)
-**
-**	All references to the word "star" were replaced by "percent"
-**	All references to the word "wild" were replaced by "like"
-**
-**	All the nice shell RE matching stuff was replaced by just "_" and "%"
-**
-**	As I don't have a copy of the SQL standard handy I wasn't sure whether
-**	to leave in the '\' escape character handling.
-**
-**	Keith Parks. <keith@mtcc.demon.co.uk>
-**
-**	SQL92 lets you specify the escape character by saying
-**	LIKE <pattern> ESCAPE <escape character>. We are a small operation
-**	so we force you to use '\'. - ay 7/95
-**
-**	Now we have the like_escape() function that converts patterns with
-**	any specified escape character (or none at all) to the internal
-**	default escape character, which is still '\'. - tgl 9/2000
-**
-** The code is rewritten to avoid requiring null-terminated strings,
-** which in turn allows us to leave out some memcpy() operations.
-** This code should be faster and take less memory, but no promises...
-** - thomas 2000-08-06
-**
-*/
-
-
-/*--------------------
- *	Match text and p, return LIKE_TRUE, LIKE_FALSE, or LIKE_ABORT.
- *
- *	LIKE_TRUE: they match
- *	LIKE_FALSE: they don't match
- *	LIKE_ABORT: not only don't they match, but the text is too short.
- *
- * If LIKE_ABORT is returned, then no suffix of the text can match the
- * pattern either, so an upper-level % scan can stop scanning now.
- *--------------------
+ * Same as above, but specifically for bytea (binary) datatype
  */
-
 static int
-MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
+MatchBytea(unsigned char *t, int tlen, unsigned char *p, int plen)
 {
 	/* Fast path for match-everything pattern */
 	if ((plen == 1) && (*p == '%'))
@@ -454,8 +510,8 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 		if (*p == '\\')
 		{
 			/* Next pattern char must match literally, whatever it is */
-			NextChar(p, plen);
-			if ((plen <= 0) || !CHAREQ(t, p))
+			BYTEA_NextChar(p, plen);
+			if ((plen <= 0) || !BYTEA_CHAREQ(t, p))
 				return LIKE_FALSE;
 		}
 		else if (*p == '%')
@@ -463,7 +519,7 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 			/* %% is the same as % according to the SQL standard */
 			/* Advance past all %'s */
 			while ((plen > 0) && (*p == '%'))
-				NextChar(p, plen);
+				BYTEA_NextChar(p, plen);
 			/* Trailing percent matches everything. */
 			if (plen <= 0)
 				return LIKE_TRUE;
@@ -474,20 +530,19 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 			 */
 			while (tlen > 0)
 			{
-
 				/*
 				 * Optimization to prevent most recursion: don't recurse
 				 * unless first pattern char might match this text char.
 				 */
-				if (CHAREQ(t, p) || (*p == '\\') || (*p == '_'))
+				if (BYTEA_CHAREQ(t, p) || (*p == '\\') || (*p == '_'))
 				{
-					int			matched = MatchText(t, tlen, p, plen);
+					int			matched = MatchBytea(t, tlen, p, plen);
 
 					if (matched != LIKE_FALSE)
 						return matched; /* TRUE or ABORT */
 				}
 
-				NextChar(t, tlen);
+				BYTEA_NextChar(t, tlen);
 			}
 
 			/*
@@ -496,9 +551,8 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 			 */
 			return LIKE_ABORT;
 		}
-		else if ((*p != '_') && !CHAREQ(t, p))
+		else if ((*p != '_') && !BYTEA_CHAREQ(t, p))
 		{
-
 			/*
 			 * Not the single-character wildcard and no explicit match?
 			 * Then time to quit...
@@ -506,8 +560,8 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 			return LIKE_FALSE;
 		}
 
-		NextChar(t, tlen);
-		NextChar(p, plen);
+		BYTEA_NextChar(t, tlen);
+		BYTEA_NextChar(p, plen);
 	}
 
 	if (tlen > 0)
@@ -516,7 +570,7 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 	/* End of input string.  Do we have matching pattern remaining? */
 	while ((plen > 0) && (*p == '%'))	/* allow multiple %'s at end of
 										 * pattern */
-		NextChar(p, plen);
+		BYTEA_NextChar(p, plen);
 	if (plen <= 0)
 		return LIKE_TRUE;
 
@@ -525,92 +579,4 @@ MatchText(unsigned char *t, int tlen, unsigned char *p, int plen)
 	 * start matching this pattern.
 	 */
 	return LIKE_ABORT;
-}	/* MatchText() */
-
-/*
- * Same as above, but ignore case
- */
-static int
-MatchTextIC(unsigned char *t, int tlen, unsigned char *p, int plen)
-{
-	/* Fast path for match-everything pattern */
-	if ((plen == 1) && (*p == '%'))
-		return LIKE_TRUE;
-
-	while ((tlen > 0) && (plen > 0))
-	{
-		if (*p == '\\')
-		{
-			/* Next pattern char must match literally, whatever it is */
-			NextChar(p, plen);
-			if ((plen <= 0) || !ICHAREQ(t, p))
-				return LIKE_FALSE;
-		}
-		else if (*p == '%')
-		{
-			/* %% is the same as % according to the SQL standard */
-			/* Advance past all %'s */
-			while ((plen > 0) && (*p == '%'))
-				NextChar(p, plen);
-			/* Trailing percent matches everything. */
-			if (plen <= 0)
-				return LIKE_TRUE;
-
-			/*
-			 * Otherwise, scan for a text position at which we can match
-			 * the rest of the pattern.
-			 */
-			while (tlen > 0)
-			{
-
-				/*
-				 * Optimization to prevent most recursion: don't recurse
-				 * unless first pattern char might match this text char.
-				 */
-				if (ICHAREQ(t, p) || (*p == '\\') || (*p == '_'))
-				{
-					int			matched = MatchTextIC(t, tlen, p, plen);
-
-					if (matched != LIKE_FALSE)
-						return matched; /* TRUE or ABORT */
-				}
-
-				NextChar(t, tlen);
-			}
-
-			/*
-			 * End of text with no match, so no point in trying later
-			 * places to start matching this pattern.
-			 */
-			return LIKE_ABORT;
-		}
-		else if ((*p != '_') && !ICHAREQ(t, p))
-		{
-
-			/*
-			 * Not the single-character wildcard and no explicit match?
-			 * Then time to quit...
-			 */
-			return LIKE_FALSE;
-		}
-
-		NextChar(t, tlen);
-		NextChar(p, plen);
-	}
-
-	if (tlen > 0)
-		return LIKE_FALSE;		/* end of pattern, but not of text */
-
-	/* End of input string.  Do we have matching pattern remaining? */
-	while ((plen > 0) && (*p == '%'))	/* allow multiple %'s at end of
-										 * pattern */
-		NextChar(p, plen);
-	if (plen <= 0)
-		return LIKE_TRUE;
-
-	/*
-	 * End of text with no match, so no point in trying later places to
-	 * start matching this pattern.
-	 */
-	return LIKE_ABORT;
-}	/* MatchTextIC() */
+}	/* MatchBytea() */
